from typing import Any, Dict
import json
import time

import numpy as np
import pandas as pd
from enums.ProcessStatusEnum import ProcessStatusEnum
from enums.DatabaseTypeEnum import DatabaseTypeEnum
from components.Utils import select_multiple, generateModel
from components.DatabaseOperation.DatabaseExtraction import get_iterator, get_steps
from components.DatabaseOperation.DatabaseLoad import save_from_dataframe
from components.DatabaseOperation.DatabaseTransformation import delete_table, delete_procedure
from components.Control import get_tasks_from_control, update_new_process
from components.S3Route import load_control_to_s3
from components.Xcom import delete_all_xcom_tasks, delete_task_instances
from enums.OperationTypeEnum import OperationTypeEnum
from components.Timezone import datetime_by_tzone

from airflow.utils.task_group import TaskGroup
from airflow.operators.python import PythonOperator
from airflow.models import Variable
from airflow.decorators import task
from airflow.exceptions import AirflowSkipException
import logging

logger = logging.getLogger()


def validate_extractor(control_params: Dict[str, Any], timezone: str, provider: str, **kwargs) -> None:
    delete_task_instances()
    ti = kwargs["ti"]
    success_tasks = ti.xcom_pull(task_ids="EXTRACTORS", key="SUCCESS_TASKS")
    failed_tasks = ti.xcom_pull(task_ids="EXTRACTORS", key="FAILED_TASKS")
    conf = ti.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="CONTROL-CONFIG")
    created_tables = ti.xcom_pull(task_ids="EXTRACTORS", key="TABLES_CREATED")
    created_tables_master = ti.xcom_pull(task_ids="MASTER_EXTRACTOR", key="TABLES_CREATED")
    lista = []
    if created_tables:
        for i in created_tables:
            lista.append(i)
    if created_tables_master:
        for i in created_tables_master:
            lista.append(i)
    final_dict = {}
    status = ProcessStatusEnum.SUCCESS.value
    if not isinstance(success_tasks, type(None)) and len(success_tasks) > 0:
        for success_task in success_tasks:
            task = ti.xcom_pull(task_ids="EXTRACTORS", key=success_task)[0]
            final_dict.update({success_task: task})
    if not isinstance(failed_tasks, type(None)) and len(failed_tasks) > 0:
        status = ProcessStatusEnum.FAIL.value
        for failed_task in failed_tasks:
            task = ti.xcom_pull(task_ids="EXTRACTORS", key=failed_task)[0]
            final_dict.update({failed_task: task})
    conf = update_new_process(conf, status, final_dict, timezone, ti, lista)
    ti.xcom_push(key="TABLES_CREATED", value=lista)
    if status == ProcessStatusEnum.FAIL.value:
        conn = control_params["connection_id"]
        bucket = control_params["bucket"]
        prefix = control_params["prefix"]
        if not prefix.endswith("/"):
            prefix += "/"
        key = prefix + control_params["filename"]
        conf = json.dumps(conf, indent=2, default=str)
        loaded = load_control_to_s3(bytes(conf.encode()), conn, bucket, key, provider, timezone)
        if loaded:
            logger.info(f"Cargado correctamente el archivo de control en {key}")
        delete_all_xcom_tasks()
        raise AirflowSkipException(f"Ocurrieron errores en la etapa de extracción")
    elif status == ProcessStatusEnum.SUCCESS.value:
        ti.xcom_push(key="CONTROL-CONFIG", value=conf)


def on_failure_extractor(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    selects = Variable.get('SELECTS', default_var=[], deserialize_json=True)
    command = selects[ti.map_index]
    tablename = select_multiple(command[1])["tablename"]
    exception = str(context["exception"])
    status = ProcessStatusEnum.FAIL.value
    init_process = ti.xcom_pull(task_ids="EXTRACTORS", key="INIT_PROCESS_DATETIME_" + str(ti.map_index))[0]
    task_result = {"description": tablename, "status": status, "message": exception, "init_process": init_process,
                   "end_process": ""}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="FAILED_TASKS", value=task_name)


def on_success_extractor(context) -> None:
    ti = context["ti"]
    task_name = f"{ti.task_id}_{ti.map_index}"
    selects = Variable.get('SELECTS', default_var=[], deserialize_json=True)
    if len(selects) == 0:
        raise AirflowSkipException(f"No se encontraron operaciones de extraccion")
    command = selects[ti.map_index]
    tablename = select_multiple(command[1])["tablename"]
    status = ProcessStatusEnum.SUCCESS.value
    init_process = ti.xcom_pull(task_ids="EXTRACTORS", key="INIT_PROCESS_DATETIME_"+str(ti.map_index))[0]
    end_process = ti.xcom_pull(task_ids="EXTRACTORS", key="END_PROCESS_DATETIME_" + str(ti.map_index))[0]
    task_result = {"description": tablename, "status": status, "message": "", "init_process": init_process,
                   "end_process": end_process}
    ti.xcom_push(key=task_name, value=task_result)
    ti.xcom_push(key="SUCCESS_TASKS", value=task_name)


def extract_from_source(command, source_conn, intern_conn, chunksize: int, timezone: str, **kwargs):

    if isinstance(command, type(None)):
        raise AirflowSkipException
    extract_tables = []
    task = kwargs['ti']
    created_tables = task.xcom_pull(task_ids="MASTER_EXTRACTOR", key="TABLES_CREATED")
    for i in created_tables:
        extract_tables.append(i)
    init_process_datetime = datetime_by_tzone(timezone).strftime('%d/%m/%Y %H:%M:%S')
    task.xcom_push(key="INIT_PROCESS_DATETIME_" + str(task.map_index), value=init_process_datetime)
    extract_type = command[0].split("|")[0]
    command = command[1]
    source_engine = source_conn.engine
    definitions = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="EXTRACTION-DEFINITION-JSON")
    multiple = select_multiple(command)
    tablename = multiple["tablename"]
    is_procedure = False
    if extract_type == "PROCEDURE":
        is_procedure = True
    columns_name = []
    # Create the model with the procedure descriptor
    model = None
    for procedure in definitions:
        if procedure["identifier"] != tablename:
            continue
        indexes = []
        if "indexes" in procedure.keys():
            indexes = procedure["indexes"]
        response_model = generateModel(tablename, procedure["fields"], indexes, intern_conn.db_type)
        if response_model["error"] != "":
            raise AssertionError(response_model["error"])
        model = response_model["model"]
        columns_name = [field["name"] for field in procedure["fields"]]
    if isinstance(model, type(None)):
        raise AssertionError(f"Definición del extracción para {tablename} en el json-descriptor no encontraddo")
    try:
        connection = intern_conn.engine.connect()
        result = bool(intern_conn.check_table(tablename, connection))
        resultado = intern_conn.verify_table(tablename, connection)
        if not result or not resultado:
            _ = delete_table(tablename, intern_conn)
            create = intern_conn.create_table(model)
            if create:
                logger.info(f"Creado correctamente la tabla {tablename}. Creado?: {create}")
            else:
                raise AssertionError(f"Error creando tabla {tablename}")
            if is_procedure:
                logger.info(f"procedure: {tablename}")
                command = command[len(tablename+"|"):]
                temp_connection = source_conn.get_basic_connection()
                command = source_conn.generate_sql_procedure(command)
                if source_conn.db_type == DatabaseTypeEnum.ORACLE.value:
                    cursor = temp_connection.cursor()
                    cursor.execute(command)
                    for resultSet in cursor.getimplicitresults():
                        data = []
                        for row in resultSet:
                            data.append(row)
                            if len(data) == chunksize:
                                dataframe = pd.DataFrame(data, columns=columns_name)
                                save = save_from_dataframe(dataframe, tablename, intern_conn)
                                if save:
                                    logger.debug(f"Guardado correctamente dataframe. Procesando más bloques")
                                data.clear()
                        if len(data) > 0:
                            dataframe = pd.DataFrame(data, columns=columns_name)
                            save = save_from_dataframe(dataframe, tablename, intern_conn)
                            if save:
                                logger.debug(f"Migrado correctamente todos los datos")
                                extract_tables.append(tablename)
                            data.clear()
                elif source_conn.db_type == DatabaseTypeEnum.MYSQL.value or \
                        source_conn.db_type == DatabaseTypeEnum.POSTGRES.value:
                    cursor = temp_connection.cursor()
                    cursor.execute(command)
                    data = []
                    for row in cursor._rows:
                        data.append(row)
                        if len(data) == chunksize:
                            dataframe = pd.DataFrame(data, columns=columns_name)
                            save = save_from_dataframe(dataframe, tablename, intern_conn)
                            if save:
                                logger.debug(f"Guardado correctamente dataframe. Procesando más bloques")
                            data.clear()
                    if len(data) > 0:
                        dataframe = pd.DataFrame(data, columns=columns_name)
                        save = save_from_dataframe(dataframe, tablename, intern_conn)
                        if save:
                            logger.debug(f"Migrado correctamente todos los datos")

                        data.clear()
                logger.info("Guardado correctamente todos los datos")
                source_conn.close_basic_connection()
            else:
                if command.replace(" ", "").lower().find("|select") != -1:
                    command = command[command.lower().find("select"):]
                response_steps = get_steps(command, chunksize, source_engine)
                if response_steps["error"] != "":
                    raise AssertionError(response_steps["error"])
                steps = response_steps["steps"]
                # Traemos el iterator
                response_iterator = get_iterator(command, chunksize, source_engine)
                if response_iterator["error"] != "":
                    raise AssertionError(response_iterator["error"])
                iterator = response_iterator["iterator"]
                logger.info(f"Número de pasos para migrar datos: {steps}")
                start_time = time.time()
                for step in range(steps):
                    dataframe = next(iterator)
                    dataframe = dataframe.fillna(value=np.nan)
                    save = save_from_dataframe(dataframe, tablename, intern_conn)
                    if save:
                        if dataframe.empty:
                            raise ValueError(f"La tabla {tablename} está vacia")
                        logger.info(f"Guardado correctamente dataframe en el paso {step+1}")
                    else:
                        raise AssertionError("Tabla no creada correctamente")
                logger.info(f"Tiempo del Task de descarga de scripts: {round(time.time() - start_time, 3)} segundos")
        end_process_datetime = datetime_by_tzone(timezone).strftime('%d/%m/%Y %H:%M:%S')
        task.xcom_push(key="TABLES_CREATED", value=tablename)
        task.xcom_push(key="END_PROCESS_DATETIME_" + str(task.map_index), value=end_process_datetime)
    except Exception as e:
        delete = delete_table(tablename, intern_conn)
        if delete:
            logger.info(f"Se borró correctamente la tabla {tablename}")
        raise AssertionError(f"Error creando la tabla y migrando datos. {type(e)}. {e}")


@task(task_id="MASTER_EXTRACTOR", trigger_rule='all_success')
def get_select_from_xcom(db_intern_conn, **kwargs):
    task = kwargs['ti']
    final_selects = []
    lista = []
    tablas_temp = set()
    conf = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="CONTROL-CONFIG")
    definitions = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="EXTRACTION-DEFINITION-JSON")
    temp_conf = conf[-1]["objects_created"] if "objects_created" in conf[-1].keys() else []
    logger.info(f"OBJETOS CREADOS: {temp_conf}")
    identificadores = [item["identifier"] for item in definitions]
    if temp_conf:
        for i in temp_conf:
            for table in definitions:
                if i == table["identifier"]:
                    if "fields" not in table.keys():
                        tablas_temp.add(i)
                        break
                elif i not in identificadores:
                    tablas_temp.add(i)
                    break

        logger.info(f"tablas temporales: {tablas_temp}")
        tablas_temp = list(tablas_temp)
        if tablas_temp:
            for i in tablas_temp:
                delete = delete_table(i, db_intern_conn)
                delete2 = delete_procedure(i, db_intern_conn)
                if delete:
                    logger.info(f"Borrado correctamente la tabla temporal {i}")
                if delete2:
                    logger.info(f"Borrado correctamente procedure {i}")
    tasks = get_tasks_from_control(conf, "extractor")
    success_tasks = tasks["tasks"]
    success_tasks = [item[1] for item in success_tasks]
    logger.info(f"COMANDOS QUE FUERON EXITOSOS: {success_tasks}")
    for i in success_tasks:
        lista.append(i)
    xcom_keys = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key="XCOM-EXTRACTION-NAMES")
    logger.debug(xcom_keys)
    for key in xcom_keys:
        if not key.startswith(OperationTypeEnum.SELECT.value) and not key.startswith(OperationTypeEnum.PROCEDURE.value):
            continue
        xcom_selects = task.xcom_pull(task_ids="SCRIPTS-EXTRACTOR", key=key)
        logger.info(f"Trayendo comandos {xcom_selects}")
        for select in xcom_selects:
            tablename = select_multiple(select)["tablename"]
            if tasks["reset"] or tasks["status"] == ProcessStatusEnum.SUCCESS.value or tablename not in success_tasks:
                final_selects.append((key, select, ))
    logger.info(f"Comandos para la extracción: {final_selects}")
    Variable.set(key='SELECTS', value=final_selects, serialize_json=True)
    logger.info(f"TABLAS CREADAS MASTER EXT: {lista}")
    task.xcom_push(key="TABLES_CREATED", value=lista)
    if len(final_selects) > 0:
        return [[item] for item in final_selects]
    else:
        return [[None]]


def get_extract_task_group(db_source_conn, db_intern_conn, chunksize: int, timezone: str,
                           control_s3: Dict[str, Any], provider: str) -> TaskGroup or None:
    group = None
    try:
        with TaskGroup(group_id="ExtraccionDeDatos", prefix_group_id=False) as group:
            selects = get_select_from_xcom(db_intern_conn)

            validate_task = PythonOperator(
                task_id="VALIDATE_EXTRACTION",
                python_callable=validate_extractor,
                op_kwargs={'control_params': control_s3, 'timezone': timezone, 'provider': provider},
                trigger_rule='all_done'
            )
            tasks = PythonOperator.partial(
                task_id="EXTRACTORS",
                python_callable=extract_from_source,
                op_kwargs={'source_conn': db_source_conn, 'intern_conn': db_intern_conn, 'chunksize': chunksize,
                           'timezone': timezone},
                on_failure_callback=on_failure_extractor,
                on_success_callback=on_success_extractor
            ).expand(op_args=selects)

            selects >> tasks >> validate_task
    except Exception as e:
        logger.error(f"Error creando taskGroup de extracción. {e}")
    finally:
        return group

