from typing import List, Any, Dict, Tuple
import uuid
import os
import shutil
import pandas as pd
from sqlalchemy import Column
from sqlalchemy import Index
from sqlalchemy.exc import InvalidRequestError
from enums.CatalogConfigurationEnum import CatalogConfigurationEnum
from enums.FileTypeEnum import FileTypeEnum
from enums.DataTypeEnum import DataTypeEnum
from enums.DataTypeOrmEnum import DataTypeOrmEnum
from enums.DatabaseTypeEnum import DatabaseTypeEnum
from components.Model.InsumoModel import InsumoModel, InsumoModelOracle
from enums.CommentsScriptEnum import CommentsScriptEnum
from components.Timezone import datetime_by_tzone

import logging

logger = logging.getLogger()


def get_type_file(key: str) -> FileTypeEnum:
    result = FileTypeEnum.EXCEL
    try:
        file_type_sufix = key.rfind(".")
        file_type = key[file_type_sufix + 1:]
        result = FileTypeEnum(file_type)
    except Exception as e:
        logger.error(f"Error obteniedo el tipo de archivo de {key}. {e}")
    return result


def get_modified_prefix(prefix: str) -> str:
    try:
        if prefix == "/":
            prefix = ""
        elif not prefix.endswith("/") and prefix != "":
            prefix += "/"
    except Exception as e:
        logger.error(f"Error modificando prefijo de {prefix}. {e}")
    finally:
        return prefix


def add_period_to_sufix(name: str, period: str) -> str:
    result = name
    try:
        position = name.rfind(".")
        result = name[:position] + period + name[position:]
    except Exception as e:
        logger.error(f"Error añadiendo periodo al nombre del archivo {name}. {e}")
    return result


def remove_invalid_rows(df: pd.DataFrame, valid_cols: List[str]) -> pd.DataFrame:
    try:
        df = df.dropna(how='all', subset=valid_cols)
    except Exception as e:
        logger.error(f"Error removiendo filas inválidas. {e}")
    finally:
        return df


def remove_fields(df: pd.DataFrame, fields_omitted: List[str]) -> pd.DataFrame:
    try:
        if len(fields_omitted) > 0:
            df = df.loc[:, ~df.columns.isin(fields_omitted)]
    except Exception as e:
        logger.error(f"Error removiendo columnas. {e}")
    finally:
        return df


def update_dict_with_catalogs(data_dict: Dict[str, Any], data: Dict[str, Any], catalog_name: str,
                              default_prefix: str) -> Dict[str, Any]:
    try:
        catalog = data[catalog_name]
        catalog_type = catalog["type"]
        catalog_prefix = catalog["prefix"]
        if catalog_type == CatalogConfigurationEnum.CATALOGO.value:
            catalog_prefix = get_modified_prefix(catalog_prefix)
        else:
            catalog_prefix = default_prefix
        s3_catalog = catalog_prefix + catalog["pattern"]
        data_dict.update({'s3_' + catalog_name: s3_catalog, catalog_name + '_key': catalog["key_field"],
                          catalog_name + '_value': catalog["value_field"]})
        if "delimiter" in catalog.keys():
            data_dict.update({catalog_name + '_delimiter': catalog["delimiter"]})
    except Exception as e:
        raise AssertionError(f"Error actualizando dict de catalogos. {e}")
    finally:
        return data_dict


def update_sql_commands_2(dataset: List[Tuple[str, str]], label_tablename: str,
                          extraction_mask: str) -> List[Tuple[str, List[str]]]:
    result = []
    comments = [CommentsScriptEnum[item].value for item in CommentsScriptEnum._member_names_ if item != CommentsScriptEnum.EXTENDED.name]
    try:
        for row in dataset:
            data = row[1].split("\n")
            data = [item.replace("\r", "") for item in data if item.strip() != '']
            final_data = []
            start_sentence = True
            add_next = False
            final_command = ""
            tablename = ""
            extend_comment = False
            for item in data:
                if not extend_comment and item.strip().startswith(CommentsScriptEnum.EXTENDED.value):
                    extend_comment = True
                    continue
                if extend_comment and item.strip().startswith(CommentsScriptEnum.EXTENDED.value):
                    extend_comment = False
                    continue
                if extend_comment:
                    continue
                if any([item.strip().startswith(comment) for comment in comments]):
                    if item.find(label_tablename+":") != -1:
                        index = item.find(label_tablename+":")
                        label_lenght = len(label_tablename+":")
                        tablename = item[index+label_lenght:].strip()
                        add_next = True
                else:
                    if start_sentence:
                        final_command = item
                    else:
                        final_command += " " + item
                    if final_command.strip().endswith(";"):
                        start_sentence = True
                        if add_next:
                            final_command = tablename + "|" + final_command.strip()[:-1]
                            add_next = False
                        final_item = final_command.replace("%", "%%")
                        final_data.append(final_item)
                    else:
                        start_sentence = False
            result.append((row[0], final_data))
            logger.info(f"Lista de comandos: {result}")
    except Exception as e:
        logger.error(f"Error extrayendo comandos sql. {e}")
    finally:
        return result


def update_sql_commands(dataset: List[Tuple[str, str]], label_tablename: str,
                        extraction_mask: str) -> List[Tuple[str, List[str]]]:
    """
    DEPRECATED: This method is replaced by update_sql_commands_2
    :param dataset:
    :param label_tablename:
    :param extraction_mask:
    :return:
    """
    result = []
    allowed_commands = ["create", "update", "delete", "select", "alter", "drop", "begin", "commit"]
    comments = [CommentsScriptEnum[item].value for item in CommentsScriptEnum._member_names_]
    try:
        for row in dataset:
            data = row[1].split(";")
            data = [item.replace("\r", "").replace(";", "").replace("\n", " ") for item in data]
            add_next = False
            final_data = []
            table_name = ""
            for item in data:
                if item.lower().strip() == "end":
                    final_data[-1] = final_data[-1] + "; end;"

                final_item = item
                if item.lower().strip().find(label_tablename.lower().strip() + ":") != -1:
                    init_index = item.replace(" ", "").lower().strip().index(label_tablename.lower().strip() + ":")
                    table_name = item.replace(" ", "").strip()[init_index + len(label_tablename + ":"):].strip()
                    add_next = True
                elif item != "":
                    if add_next:
                        item = table_name + "|" + item
                        add_next = False
                    final_item = item.strip()
                    table_name = ""
                if any([command in final_item.lower() for command in allowed_commands]):
                    if final_item.strip()[:2] in comments:
                        init_indexes = [final_item.lower().find(command) for command in allowed_commands
                                        if final_item.lower().find(command) != -1]
                        if len(init_indexes) > 0:
                            init_index = init_indexes[0]
                            if row[0].lower().find(extraction_mask.lower()) != -1 and \
                                final_item.replace(" ", "").startswith(CommentsScriptEnum.DASHES.value) and \
                                    not final_item.replace(" ", "")[len(CommentsScriptEnum.DASHES.value):].startswith(
                                        label_tablename):
                                continue
                            final_item = final_item[init_index:]
                final_item = final_item.replace("%", "%%")
                final_data.append(final_item)
            final_data = [item.replace("\t", "") for item in final_data if item != "" and
                          all([not item.strip().startswith(comment) for comment in comments])]
            result.append((row[0], final_data))
            logger.info(f"Lista de comandos: {result}")
    except Exception as e:
        logger.error(f"Error extrayendo comandos sql. {e}")
    finally:
        return result


def select_multiple(command: str) -> Dict[str, Any]:
    response = {'is_multiple': False, 'tablename': ''}
    tablename = ""
    no_procedure_init = "|select"
    try:
        if command.lower().replace(" ", "").find(no_procedure_init) == -1:
            response["is_multiple"] = True
            tablename = command[:command.index("|")].strip()
            response["tablename"] = tablename
        else:
            if command.lower().replace(" ", "").find(no_procedure_init) != -1:
                response["is_multiple"] = True
                tablename = command[:command.index("|")].strip()
            response["tablename"] = tablename
    except Exception as e:
        raise AssertionError(f"Error validando si es múltiple select y nombre de tabla. {e}")
    finally:
        return response


def create_temp_file(tmp_path: str, filename_mask: str, file_type: str, tablename: str, timezone: str,
                     pattern: str) -> str:
    """ Create an output result as a file with a mask in the name
    """
    fullpath = ""
    try:
        dir_name = str(uuid.uuid4())
        if not tmp_path.endswith("/"):
            tmp_path += "/"
        path_dir = tmp_path + dir_name
        os.makedirs(path_dir)
        current_datetime = str(datetime_by_tzone(timezone, pattern))
        filename_mask = filename_mask.replace("<source_name>", tablename)
        filename_mask = filename_mask.replace("<datetime>", current_datetime)
        filename = filename_mask + "." + file_type
        fullpath = path_dir + "/" + filename
        open(fullpath, mode='a').close()
    except Exception as e:
        logger.error(f"Error creando directorio y archivo temporal.{e}")
    finally:
        return fullpath


def delete_temp_dir(module_name: str) -> bool:
    drop = False
    try:
        if os.path.exists(module_name):
            directory = os.path.dirname(module_name)
            shutil.rmtree(directory, ignore_errors=True)
    except Exception as e:
        raise AssertionError(f"Error borrando modulo temporal. {e}")
    finally:
        return drop


def generateModel(tablename: str, attributes: List[Dict[str, Any]], indexes: List[Dict[str, Any]], db_target: str,
                  modelName: str = "TableModel") -> Dict[str, Any]:
    response = {'model': None, 'error': ''}
    default_precision = 8
    model_args = {
        '__tablename__': tablename,
        '__table_args__': {'extend_existing': True}
    }
    if db_target == DatabaseTypeEnum.ORACLE.value:
        model = type(modelName, (InsumoModelOracle,), model_args)
    else:
        model = type(modelName, (InsumoModel,), model_args)
    try:
        for attribute in attributes:
            if attribute["datatype"] == DataTypeEnum.TEXT.name and "maxLength" in attribute.keys():
                setattr(model, attribute["name"],
                        Column(DataTypeOrmEnum[attribute["datatype"]].value(attribute["maxLength"])))
            elif attribute["datatype"] == DataTypeEnum.DECIMAL.name:
                precision = default_precision
                if "decimal_precision" in attribute.keys():
                    precision = attribute["decimal_precision"]
                setattr(model, attribute["name"],
                        Column(DataTypeOrmEnum[attribute["datatype"]].value(38, precision)))
            else:
                setattr(model, attribute["name"],
                        Column(DataTypeOrmEnum[attribute["datatype"]].value))
        model = model.__table__
        for current_index in indexes:
            index_name = tablename + "_" + current_index["name"]
            fields = current_index["index_fields"]
            final_fields = [model.c[field] for field in fields]
            Index(index_name, *final_fields)
        response["model"] = model
    except InvalidRequestError as e:
        message = f"InvalidRequestError. {e}"
        logger.debug(message)
        response["error"] = message
    except Exception as e:
        message = f"Error creando modelo dinámico. {e}"
        logger.error(message)
        response["error"] = message
    finally:
        return response


def generate_indexes(tablename: str, indexes: List[Dict[str, Any]], db_target: str) -> None:
    try:
        for current_index in indexes:
            index_name = current_index["name"]
            fields = current_index["index_fields"]

    except Exception as e:
        logger.error(f"Error creando índices. {e}")


def delete_temp_dirs(tmp_dir: str) -> bool:
    delete = False
    try:
        dirs = list(os.listdir(tmp_dir))
        for directory in dirs:
            full_path = tmp_dir + "/" + directory
            if os.path.isdir(full_path):
                _ = delete_temp_dir(full_path)
                logger.debug(f"Se borró el directorio {directory}")
        if not os.path.exists(tmp_dir):
            os.makedirs(tmp_dir)
        delete = True
    except Exception as e:
        logger.error(f"Error borrando archivos temporales en {tmp_dir}. {e}")
    finally:
        return delete
